package polars

  1. Overview
  2. Docs

polars-ocaml

🐻‍❄️❤️🐫

polars-ocaml is a project to provide idiomatic OCaml bindings to the Polars dataframe library.

Check out the examples from the Polars user guide for a quick tour of what you can do with polars-ocaml!

project status

build

Currently very much WIP. Please expect breakages and large changes in API.

Note that the current code assumes that the OCaml version is 4.14.1; see HACKING.md for notes on how to build polars-ocaml.

utop

polars-ocaml works in utop!

$ dune utop --profile utop^utop-workaround

polars-ocaml running in utop

license

This project is licensed under the terms of the MIT license, with the exception of the following files: